#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This file is part of tof-simulator.
#
# tof-simulator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tof-simulator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tof-simulator.  If not, see <https://www.gnu.org/licenses/>.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk3agg import FigureCanvasGTK3Agg as FigureCanvas
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3 as NavigationToolbar
# from matplotlib.widgets import Cursor
import sys
import os
import subprocess as sub
from .version import VERSION, COPYRIGHT
from .tof import ToF
from .nist_elem import analyze_substance

default_conffile = os.path.join(os.path.dirname(__file__), 'tof.conf')


class tof_gtk:
  Npoints_fast = 100000

  def __init__(self, conffile=None):

    # use GtkBuilder to build our interface from the XML file
    try:
      builder = Gtk.Builder()
      interface = os.path.join(os.path.dirname(os.path.realpath(__file__)), "tof_gtk.xml")
      builder.add_from_file(interface)
    except BaseException:
      self._message("Failed to load UI XML file: tof_gtk.xml")
      sys.exit(1)

    if conffile is None: conffile = default_conffile
    self.masses = self.init_calc(conffile)
    self.init_gui(builder)
    self.init_graph()
    self.line = None

    # connect signals
    builder.connect_signals(self)
    self.mw.set_default_size(1000, 600)
    # self.Npoints = self.tof.Npoints       # Usado para guardar el valor seteado por user

  def _message(self, message):
    # log to terminal window
    print(message)

  # Create the tof object
  def init_calc(self, conffile=None):
    self.tof = ToF()
    return self.tof.load_conf_file(conffile)

  def init_gui(self, builder):
    # get the widgets which will be referenced in callbacks
    self.mw = builder.get_object('w1')   # Main window
    self.sw1 = builder.get_object('sw1')   # subwindow with plot
    self.sw2 = builder.get_object('sw2')   # subwindow with toolbar
    self.tv = builder.get_object('treeview1')
    self.lm = builder.get_object('listmass')   # List of Masses

    self.addmass = builder.get_object('addmasas')
    self.enmass = builder.get_object('enmass')

    self.swupd = builder.get_object('swfast')   # Auto-update
    self.isotop = builder.get_object('chiso')   # Get isotopes

    self.popup_menu = builder.get_object('tvmenu')
    self.cbaddrow = builder.get_object('cbaddrow')
    # Preferences window
    self.preferences_dialog = builder.get_object('dialog_pref')
    self.sbs = builder.get_object('sb_s')
    self.sbd = builder.get_object('sb_d')
    self.sbD = builder.get_object('sb_D')
    self.sbLt = builder.get_object('sbLt')
    self.cmLt = builder.get_object('comboLt')
    self.sbLr = builder.get_object('sbLr')
    self.cmLr = builder.get_object('comboLr')
    self.enNp = builder.get_object('enNpoints')
    self.entp = builder.get_object('entimeprec')
    self.chneg = builder.get_object('chneg')
    self.chgrd = builder.get_object('chgrd')
    self.threshold_isotope = builder.get_object('thsubst')

    self.sbEs = builder.get_object('sbEs1')   # Potencial Vs
    self.sbEd = builder.get_object('sbEd1')   # Potencial Vd
    self.sbT = builder.get_object('sbT')      # Temperature

    self.update_display_tof()

    # Initialize the masses with default values
    if self.masses is None:
      self.masses = {
          '235UF6': {'M': 349, 'P': 0.03, 'tex': r"$^{{235}}UF_{{6}}$"},
          '238UF6': {'M': 352, 'P': 0.37, 'tex': r"$^{{238}}UF_{{6}}$"},
          '235UF5': {'M': 328, 'P': 0.07, 'tex': r"$^{{235}}UF_{{5}}$"},
          '238UF5': {'M': 333, 'P': 0.53, 'tex': r"$^{{238}}UF_{{5}}$"},
      }
    self.update_display_mass()

  def update_display_tof(self):
    # Initialize the values in the gui
    self.sbEs.set_value(self.tof.Vs)
    self.sbEd.set_value(self.tof.Vd)
    self.sbT.set_value(self.tof.Temperature)
    self.sbs.set_value(self.tof.s)
    self.sbd.set_value(self.tof.d)
    self.sbD.set_value(self.tof.D)
    self.sbLt.set_value(self.tof.dt * 1.e3)
    self.sbLr.set_value(self.tof.ds * 10)
    self.enNp.set_text(str(self.tof.Npoints * 1.e-6))
    self.entp.set_text(str(self.tof.timeprec * 1.e3))
    if self.tof.tdist.lower() == 'normal': self.cmLt.set_active(0)
    else: self.cmLt.set_active(1)
    if self.tof.posdist.lower() == 'normal': self.cmLr.set_active(0)
    else: self.cmLr.set_active(1)

  def update_display_mass(self):
    self.lm.clear()
    for k, v in self.masses.items():
      # if v.has_key('L'): key= v['L']
      # else:             key=k
      self.lm.append([k, v['M'], v['P']])

  # def onclick(self, event):
  #   print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
  #     event.button, event.x, event.y, event.xdata, event.ydata)
  #   x= [event.xdata,event.xdata]
  #   y= self.ax.get_ylim()
  #   if self.line == None:
  #     self.line, = self.ax.plot(x,y,'r',lw=2)
  #   else:
  #     self.line.set_xdata(x)
  #     self.line.set_ydata(y)
  #   #We need to draw
  #   self.canvas.draw()

  # Create the graph and add to the window
  def init_graph(self):
    self.fig = Figure(figsize=(5, 5), dpi=100)
    self.ax = self.fig.add_subplot(111)

    self.canvas = FigureCanvas(self.fig)
    self.canvas.set_size_request(400, 400)
    self.sw1.add_with_viewport(self.canvas)

    toolbar = NavigationToolbar(self.canvas, self.mw)
    self.sw2.add_with_viewport(toolbar)

  def graph(self):
    species, X = self.tof.data_to_array()
    if self.chneg.get_active():
      X[1:] *= -1
    self.ax.cla()

    if species is not None:
      # for i,k in enumerate(sorted(species[1:-1], key= lambda x: self.masses[x]['M'])):
      for i, k in enumerate(species[1:-1]):
        lab = self.masses[k].get('L', k)
        self.ax.plot(X[0, :], X[i + 1, :], lw=3, label=lab)
      self.ax.plot(X[0, :], X[-1, :], 'o-k', ms=2, lw=1)

      ymin, ymax = X[-1, :].min(), X[-1, :].max()
      self.ax.set_ylim((ymin * 1.05, 1.05 * ymax))
      self.ax.set_xlabel(r" $t(\mu s)$", fontsize="x-large")
      self.ax.set_ylabel(r"$\rho(t)$", fontsize='x-large')
      ncol = (len(species) - 2) // 8 + 1
      self.ax.legend(title=r'$T={0}^{{\circ}}K$'.format(self.tof.Temperature),
                     loc='best', fontsize='small', framealpha=0.5, ncol=ncol)

      leyenda = u'Parámetros del TOF\n'
      for k in ['s', 'd', 'D', 'Vs', 'Vd']:
        leyenda += r'${0}={1}$  '.format(k, self.tof.__dict__[k])
      self.ax.set_title(leyenda)

    if self.chgrd.get_active():
      self.ax.grid(True)
    else:
      self.ax.grid(False)
    self.canvas.draw()

  def on_addmasas_clicked(self, widget):
    subst = self.enmass.get_text().split(',')
    thr = float(self.threshold_isotope.get_text()) * 1.e-2
    isotopes = self.isotop.get_active()
    for s in subst:
      self.masses.update(analyze_substance(s.strip(), threshold=thr, isotopes=isotopes))

    self.update_display_mass()
    self.enmass.set_text('')

  def on_swupd_activate(self, switch):
    pass
    # if switch.get_active():      self.tof.Npoints = 100000
    # else:                        self.tof.Npoints = self.Npoints

  def on_sbEd1_value_changed(self, spinbutton):
    if self.swupd.get_active():
      self.update_plot()
    # if self.swupd.get_active():
    #   self.update_plot()

  def on_sbEs1_value_changed(self, spinbutton):
    if self.swupd.get_active():
      self.update_plot()

  def on_about_activate(self, menuitem):
    authors = ["Juan Fiol - para LASIE"]
    ad = Gtk.AboutDialog()
    ad.set_transient_for(self.mw)
    ad.set_destroy_with_parent(True)
    ad.set_program_name(u"Simulación de TOF - Version {}".format(VERSION))
    ad.set_copyright("Copyright 2020 Juan Fiol\n" + COPYRIGHT)
    ad.set_comments(u"Time-of-flight mass spectrometer simulator")
    ad.set_authors(authors)
    ad.set_logo_icon_name(Gtk.STOCK_EXECUTE)
    # callbacks for destroying the dialog

    def close(dialog, response, editor):
      dialog.destroy()

    def delete_event(dialog, event, editor):
      return True
    ad.connect("response", close, self)
    ad.connect("delete-event", delete_event, self)
    ad.show()

  def on_teoria_activate(self, menuitem):
    """Open the pdf file.
    TODO: Change to html documentation, possibly online
    """
    pdfteoria = os.path.join(os.path.dirname(__file__), 'tof_teoria.pdf')
    viewers = ['evince', 'okular', 'acrobat']
    for v in viewers:
      status = sub.call([v, pdfteoria])
      if status == 0: return 0

  # Save configuration (File -> Save)
  def on_save_activate(self, menuitem):
    fcd = Gtk.FileChooserDialog("Save Configuration...", None,
                                Gtk.FileChooserAction.SAVE,
                                (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                 Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
    response = fcd.run()
    if response == Gtk.ResponseType.OK:
      self.tof.save_conf_file(fcd.get_filename(), self.masses)
    fcd.destroy()

  # Export data
  def on_export_activate(self, menuitem):
    fcd = Gtk.FileChooserDialog("Export data...", None,
                                Gtk.FileChooserAction.SAVE,
                                (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                 Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
    response = fcd.run()
    if response == Gtk.ResponseType.OK: self.tof.save_data((fcd.get_filename()))
    fcd.destroy()

  # Load configuration (File -> Open)
  def on_open_activate(self, data=None):
    fcd = Gtk.FileChooserDialog("Open Configuration...",
                                None,
                                Gtk.FileChooserAction.OPEN,
                                (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                 Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
    response = fcd.run()
    if response == Gtk.ResponseType.OK:
      self.load_conf_file(fcd.get_filename())
    fcd.destroy()

  def load_conf_file(self, fname):
    m = self.tof.load_conf_file(fname)
    if m is not None:
      self.masses = m
      self.update_display_mass()
    self.update_display_tof()

  # When the main window is destroyed, we want to break out of the GTK main loop.
  def on_w1_destroy(self, widget, data=None):
    self._message('Exterminate!')
    Gtk.main_quit()

  # Called when the user clicks the 'Quit' menu.
  def on_mquit_activate(self, menuitem):
    self.on_w1_destroy(menuitem)

  # Edicion de Especie, Masa y Poblacion ################
  def on_cellEsp_edited(self, widget, path, text):
    self.lm[path][0] = text

  def on_spMass_edited(self, widget, path, text):
    self.lm[path][1] = float(text.replace(',', '.'))

  def on_spPob_edited(self, widget, path, text):
    self.lm[path][2] = float(text.replace(',', '.'))

  def on_treeview1_button_press_event(self, tv, event):
    if event.button == 3:
      time = event.time
      self.popup_menu.popup(None, None, None, None, event.button, time)

  def on_addrow_activate(self, menuitem):
    # selection = self.tv.get_selection()
    # model, pathlist= selection.get_selected_rows()
    # if pathlist != []:
    #   j  = model.get_iter(pathlist[-1])
    #   self.lm.insert_after(j)
    # else:
    self.add_rows(1, -1)

  def on_butAdd_clicked(self, menuitem):
    nrows = int(self.cbaddrow.get_active_text().split()[0])
    self.add_rows(nrows, -1)

  def add_rows(self, nrows, position):
    for n in range(nrows):
      self.lm.insert(position)

  def on_butDel_clicked(self, menuitem):
    self.delete_selected_masses()

  def delete_selected_masses(self):
    selection = self.tv.get_selection()
    if selection is not None:
      model, pathlist = selection.get_selected_rows()
      for path in reversed(pathlist):
        tree_iter = model.get_iter(path)
        self.masses.pop(self.lm.get_value(tree_iter, 0))
        self.lm.remove(tree_iter)
      self.update_masas()

# #######################################################################

  # Setup del TOF
  def on_butTOFset_clicked(self, menuitem, data=None):
    self.preferences_dialog.present()

  def on_butPrefOK_clicked(self, widget):
    self.update_tof_param()

  def on_dialog_pref_delete_event(self, widget, data):
    # self.preferences_dialog.hide()
    return True

  def update_tof_param(self):
    self.tof.Temperature = self.sbT.get_value()
    self.tof.Vs = self.sbEs.get_value()
    self.tof.Vd = self.sbEd.get_value()

    self.tof.s = self.sbs.get_value()
    self.tof.d = self.sbd.get_value()
    self.tof.D = self.sbD.get_value()

    self.tof.ds = self.sbLr.get_value() * 0.1
    self.tof.dt = self.sbLt.get_value() * 1.e-3

    if self.swupd.get_active(): self.tof.Npoints = self.Npoints_fast
    else: self.tof.Npoints = int(float(self.enNp.get_text()) * 1000000)

    self.tof.timeprec = float(self.entp.get_text()) * 1.e-3

    self.tof.tdist = self.cmLt.get_active_text()
    self.tof.posdist = self.cmLr.get_active_text()

  def on_butPrefClose_clicked(self, widget):
    self.preferences_dialog.hide()

  def on_chgrd_toggled(self, widget):
    self.graph()

  def on_chneg_toggled(self, widget):
    self.graph()
# #######################################################################

  # Calculations
  def calculate(self, particles):
    return self.tof.signal(particles)

  def update_plot(self):
    self.update_tof_param()
    self.update_masas()
    self.times = self.calculate(self.masses)
    self.graph()

  def on_butCalc_clicked(self, widget):
    self.update_plot()

  def update_masas(self):
    m = {}
    for label, Mass, Pobl in self.lm:
      if label.strip() != '':
        if label in self.masses: LL = self.masses[label].get('L', label)
        else: LL = label

        m[label] = {'M': Mass, 'P': Pobl, 'L': LL}
    if m != {}:
      self.masses = m              # Update all parameters

# #######################################################################

  # Run main application window
  def main(self):
    self.mw.show_all()
    Gtk.main()


if __name__ == "__main__":
  import argparse
  from version import VERSION
  default_conffile = os.path.join(os.path.dirname(__file__), 'tof.conf')
  parser = argparse.ArgumentParser(
      description=u'"Simulación de la señal obtenida en el tiempo de vuelo"')

  parser.add_argument('-V', '--version', action='version',
                      version='%(prog)s version {}'.format(VERSION))
  parser.add_argument(
      "-c",
      "--conf",
      default=default_conffile,
      metavar='File',
      help="Configuration file to load")

  args = parser.parse_args()

  TOF_w1 = tof_gtk(args.conf)
  TOF_w1.main()



VERSION = '3.00'
COPYRIGHT = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
"""



# -*- coding: utf-8 -*-
#
# This file is part of tof-simulator.
#
# tof-simulator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tof-simulator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tof-simulator.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np
from math import sqrt
from configparser import ConfigParser
import os
from .nist_elem import Sustancias


_tof_esquema = r"""
Diagram of TOF
--------------

  |<-- 2s --->|<--- d ---->|<--------------D---------------->|
 _____________________________________________________________
  |           |            |                                 |
  |     Vs    |     Vd     |         E=0                     |
  |           v            |                                  \
  |     O  - - - - - - - - - - - - - - - - - - - - - - - - -   >~~~ Detector
  |           ^ r0         |                                  /
  |           |            |                                 |
 _|___________|____________|_________________________________|

"""

_default_param = {
    's': 1.0,                   # Distance from center to first plate in cm
    'd': 2.54,                  # Distance between plates in cm (second stage)
    'D': 105.,                  # Distance of free flight
    'Vs': 200,                  # Potential Es in eV
    'Vd': 3000,                 # Potential Ed in eV
    'ds': 0.1,                  # Radio del spot del laser in cm
    'dt': 0.008,                # Laser-pulse duration (microsegundos)
    't0': 0.,                   # Offset in time of TOF (experimental)
    'r0': 1.e3                  # Radio de abertura después de la extracción
}

_default_conditions = {
    'Temperature': 300,
    'Npoints': 1000000,      # Number of simulation particles
    'timeprec': 0.002,       # Precision de tiempo (microsegundos)
    'posdist': 'normal',     # Disribución en posición de la intensidad del láser
    'tdist': 'normal',       # Distribución en tiempos de la intensidad del láser
    'v_inicial': 0.          # Velocidad de grupo del gas en sqrt(2*eV)
}

_default_conffile = os.path.join(os.path.dirname(__file__), 'tof.conf')
if not os.path.exists(_default_conffile):
  _default_conffile = 'tof.conf'


def _get_nearest(x, x0):
  """Find index of value on x nearest to x0, for sorted x  """
  # imax = np.abs(x - x0).argmin()
  imax = np.searchsorted(x, x0, side="left")
  if x0 - x[imax] > x[imax + 1] - x0: imax += 1
  if imax > len(x): imax = len(x)
  return imax


def _get_peak(x, y, N=10):
  """Calculation of the peak position for the array y.
  Fits the proximity of the maximum with a parabole to get its maximum
  """

  j = y.argmax()

  p = np.polyfit(x[j - N:j + N], y[j - N:j + N], 2)

  x0 = np.array([np.sqrt(p[2] / p[0]), -2 * p[2] / p[1], -p[1] / (2 * p[0])]).mean()
  y0 = np.polyval(p, x0)
  imax = _get_nearest(x, x0)  # Find the closest index

  return imax, x0, y0


def get_one_peak(x, y, fwidth=1 / np.e):
  """Returns features of a single peak.
  Returns a list with:
  - Position 0: indexes of position of:

    - lower half height
    - center
    - upper half height

  - Position 1: values of x
  - Position 2: values of y
  """
  im, xm, ym = _get_peak(x, y)
  yf = fwidth * ym
  i2p = np.abs(y[im:] - yf).argmin() + im
  i2m = im - np.abs(y[im::-1] - yf).argmin()

  return [(i2m, im, i2p), (x[i2m], xm, x[i2p]), (y[i2m], ym, y[i2p])]


class ToF(object):
  """
  The ToF object defines and handles all aspects of a Time-of-Flight spectrometer

  The following parameters related to the construction and operation are included:
  Distances, voltages, working temperature, particle velocity fields and dispersion,
  time duration and size of the ionizing beam are also included.

  Examples
  --------
  ::

    tof_parameters= {
      's': 0.7,                    # Distance from center to first plate in cm
      'd': 2.54,                   # Distance between plates in cm (second stage)
      'D': 100.,                   # Distance of free flight
      'Vs': 500,                   # Potential Es in eV
      'Vd': 1900,                  # Potential Ed in eV
      'ds': 0.05,                  # Radio del spot del laser in cm
      'dt': 0.008,                 # Laser-pulse duration
      't0': 0.,                    # Offset in time of TOF (experimental)
      'r0': 1.e3,                  # Aperture radius after extraction
      }

    # We can create the object with the parameters desired
    T = ToF(['Ar', 'N2', 'CO2'],**tof_parameters)

    # We can also load parameters from a file
    T.load_conf_file('tof.conf')

    T.signal()     # Calculate the signals

    # Plot the signals
    T.make_plot(fname='tof2.png', negative=True, show_legend=True, show_all=True)

    # Just printing the object gives all the information on construction and
    # condition parameters as well as the substances being simulated
    print(T)

  """

  tof_parameters = ['s', 'd', 'D', 'Vs', 'Vd', 'ds', 'dt', 't0', 'r0']
  adic_parameters = ['Temperature', 'Npoints', 'timeprec', 'posdist',
                     'tdist', 'v_inicial']

  _esquema = _tof_esquema

  def __init__(self, substances=None, **kwds):
    """
    Initialize a ToF object defines and handles all aspects of a Time-of-Flight spectrometer

    Parameters
    ----------
    self: type
      description
    substances: str
      A string of comma-separated substances
    kwds:
      keyword arguments may include any of those defining TOF parameters o conditions. May include:

    - 's': float
          Distance from center to first plate in cm
    - 'd': float
          Distance between plates in cm (second stage)
    - 'D': float
          Distance of free flight
    - 'Vs': float
          Potential Es in eV
    - 'Vd': float
          Potential Ed in eV
    - 'ds': float
          Radio del spot del laser in cm
    - 'dt': float
          Laser-pulse duration (microsegundos)
    - 't0': float
          Offset in time of TOF (experimental)
    - 'r0': 1.e3
          Radio de abertura después de la extracción
    - 'Temperature': float
          Temperature of the gas
    - 'Npoints': intended
          Number of simulation particles
    - 'timeprec': float
          Precision in time: Time step in the spectra (microsegundos)
    - 'posdist': 'normal' or 'uniform'
          Spatial disribution of ionizing beam
    - 'tdist': 'normal' or 'uniform'
          Time distribution of ionizing pulse beam
    - 'v_inicial': float
          Group-velocity of the gas (in units of sqrt(2*eV))

    """
    # First set default values
    self.__dict__.update(_default_param)
    self.__dict__.update(_default_conditions)

    self.load_conf_file(_default_conffile)
    # Then, process input (overwrite default)
    self.__dict__.update(kwds)
    self.times = None

    if substances is not None:
      self.fragments = Sustancias(substances)
    else:
      self.fragments = Sustancias()

  def add_substances(self, substances, threshold=1.e-3):
    """Add substances to be simulated.

    Parameters
    ----------

    substances: str
       Comma-separated string, or list of strings, each substance may have the form            SF5^{+}, SF5^{++}, SF5^+, S^2+
    threshold: float
       Minimum abundance of a given isotope to be included (default 1.e-3).

    """
    self.fragments.thr = threshold  # Threshold de los isótopos a incluir
    self.fragments.add(substances)

  def remove_substances(self, sustancias):
    """Remove substances from the simulation

    Parameters
    ----------
    self: type
      description
    sustancias: str or list of strings
      substances or fragments to remove
    """
    self.fragments.remove(sustancias)

  def _message(self, message):
    # log to terminal window
    print(message)

  def load_conf_file(self, fname):
    """Load a configuration file, update tof parameters, and also return masses if present."""

    if fname is None:
      return None

    if not os.path.exists(fname):
      self._message('Configuration file {0} does not exist'.format(fname))
      return None

    parser = ConfigParser()
    parser.optionxform = str
    parser.read(fname)

    if parser.has_section('tof'):
      for name, value in parser.items('tof'):
        if name in self.tof_parameters:
          self.__dict__[name] = float(value)

    if parser.has_section('condition'):
      for name, value in parser.items('condition'):
        if name in ['Temperature', 'timeprec', 'v_inicial']:
          self.__dict__[name] = float(value)
        elif name == 'Npoints':
          self.__dict__[name] = int(value)
        elif name in ['posdist', 'tdist']:
          self.__dict__[name] = value

    sections = parser.sections()
    sections.remove('tof')
    sections.remove('condition')
    m = {}
    for section_name in sections:
      m[section_name] = {}
      for name, value in parser.items(section_name):
        try:
          m[section_name][name] = float(value)
        except BaseException:
          m[section_name][name] = value
    return m

  def save_conf_file(self, fname, masas={}):
    """Save configuration data for tof and masses to a file
    """
    parser = ConfigParser()
    parser.optionxform = str
    parser.add_section('tof')
    parser.add_section('condition')
    for opt in self.tof_parameters:
      parser.set('tof', opt, str(self.__dict__[opt]))
    for opt in self.adic_parameters:
      parser.set('condition', opt, str(self.__dict__[opt]))

    for m, values in list(masas.items()):
      parser.add_section(m)
      for k, v in list(values.items()):
        parser.set(m, k, str(v))
    fo = open(fname, 'w')
    parser.write(fo)
    fo.close()

  def get_tof_parameters(self, explain=False):
    """Returns the parameters of TOF in a dict.

    Parameters
    ----------
    self: type
      description
    explain: bool
      If `True` add an ascii graphic with the diagram of the TOF

    """
    p = {}
    for k in self.tof_parameters:
      p[k] = self.__dict__[k]
    if explain:
      return p, self._tof_esquema
    return p

  def __str__(self):
    s = """
Time-of-Flight
**************

Construction parameters:
"""
    s += self._esquema
    for k in self.tof_parameters:
      s += "  {} = {}\n".format(k, self.__dict__[k])

    s += "\n\nAditional parameters:\n\n"
    for k in self.adic_parameters:
      s += "  {} = {}\n".format(k, self.__dict__[k])

    s += "\n\nFragments detected:\n"
    s += str(self.fragments)

    return s

  def set_initial_distribution(self):
    """Initial velocity and position distributions"""
    U0 = 0.025 * (1.e-10 + self.Temperature) / 300.  # Energy in eV  (= k_B * T)
    sigma = np.sqrt(U0)         # Velocity spread (M=1)

    # velocidades iniciales en la dirección de la extracción
    v_0 = np.random.normal(loc=self.v_inicial, scale=sigma, size=self.Npoints)
    # velocidades iniciales en la dirección perpendicular a la extracción
    v_1 = np.random.normal(loc=0, scale=sigma, size=self.Npoints)
    v_2 = np.random.normal(loc=0, scale=sigma, size=self.Npoints)
    v_perp = np.sqrt(v_1**2 + v_2**2)

    if self.posdist.lower() == 'normal':
      # Distribución de posiciones: Gaussiana
      x_0 = np.random.normal(loc=0.0, scale=self.ds / 2., size=self.Npoints)
      x_0 = np.where(np.abs(x_0) <= self.s, x_0,
                     x_0 - (x_0 // self.s) * self.s)
    else:
      # Distribución de posiciones: Uniforme
      x_0 = (2 * np.random.random_sample(self.Npoints) - 1) * self.ds

    # Keep initial velocity and position distributions
    counts, bins_i = np.histogram(v_0, bins=100, density=True)
    self.v_i = np.array([(bins_i[1:] + bins_i[:-1]) / 2., counts])
    counts, bins_i = np.histogram(x_0, bins=100, density=True)
    self.x_i = np.array([(bins_i[1:] + bins_i[:-1]) / 2., counts])
    return x_0, v_0, v_perp

  def _filter_abertura(self, x0, v0, vp):
    "Filter particles that may pass through an aberture of radius self.r0"
    t_ext = self.time_of_extract(x0, v0)
    cond = vp * t_ext <= self.r0
    return x0.compress(cond), v0.compress(cond)

  def signal(self, particles=None):
    """Evaluate the signal that would produce the particles in the ToF

    Parameters
    ----------
    particles : dict each item is a dictionary that should have
        - key: is the label of the mass
        - 'M': Mass (in AMU)
        - 'P': (float in range 0 to 1) Abundance
        - 'L': an (optional, possibly formatted) label, otherwise the key is used

    If particles is None => Usa los fragments

    Returns
    -------
    times : numpy array
      Sets the object variable self.times and also returns its value
    """
    if particles is None:
      particles = self.fragments

    x0, v0, v_p = self.set_initial_distribution()

    # Filtramos las partículas que no pueden pasar la abertura de la extracción
    x_0, v_0 = self._filter_abertura(x0, v0, v_p)

    Time = self.time_of_flight(x_0, v_0)
    self.times = self.calc_signal(particles, Time)
    # Add the theoretical mean value
    self.mean_times()
    return self.times

  def calc_signal(self, particles, Time):
    "Calculo de los histogramas para generar la señal de un grupo de partículas"

    # Incerteza en el tiempo inicial (duracion del laser)
    if self.dt > 1.e-6:
      if self.tdist.lower() == 'normal':
        Dt = np.random.normal(loc=0.0, scale=self.dt / 2., size=len(Time))
      else:
        Dt = 2 * self.dt * np.random.random(len(Time)) - self.dt
    else:
      Dt = 0.

    sm = np.sqrt([p['M'] for p in list(particles.values())])
    tmin, tmax = sm.min() * Time.min() - Dt.max(), sm.max() * Time.max() + Dt.max()
    step = (tmax - tmin) / 20.
    tmin -= step
    tmax += step
    Nbins = min(int((tmax - tmin) / self.timeprec), 1000)

    x = {}
    x['signal'] = None

    for k, p in list(particles.items()):
      hist, bins = np.histogram(sqrt(p["M"]) * Time + Dt,
                                range=(tmin, tmax), bins=Nbins, density=True)
      x[k] = p["P"] * hist
      if x['signal'] is None:
        x['signal'] = x[k].copy()  # Fallaba sin .copy()
      else:
        x['signal'] += x[k]
    x['time'] = np.array((bins[1:] + bins[:-1]) / 2. + self.t0)  # Eje x
    return x

  def mean_times(self, x0=0.):
    """Evaluate the 'main' time of each species, given as if generated in ideal conditions:

      - At time t=0
      - At rest (null initial velocity)
      - At a distance x0 from the center of the extraction plates (s = 0)

    """
    Time = self.time_of_flight(x0, self.v_inicial)
    m = self.fragments
    for k in m.ListItems:
      s = m[k]
      if s['q'] >= 1:
        s['t'] = sqrt(s['M'] / s['q']) * Time + self.t0
      else:
        s['t'] = 0

  def get_statistics_peaks(self, substances='all', fwidth=1 / np.e):
    """Find peaks

    Parameters
    ----------
    substances: 'all' or list
      - if 'all' find the peak for every fragment in the TOF
      - if substance is a list of strings, each element must be a fragment in TOF

    fwidth: float
       fraction of the maximum at which to evaluate the full width.
       For instance, fwidth = 0.5 corresponds to FWHM
    """
    if substances == 'all': frags = list(self.fragments.ListItems)
    elif substances in self.fragments.ListItems: frags = [substances]
    else: frags = []

    peaks = Peaks()
    for m in frags:
      ind, x, y = get_one_peak(self.times['time'], self.times[m], fwidth=fwidth)
      peaks[m] = {'index': ind, 'position': x[1], 'height': y[1], 'width': x[2] - x[0]}

    return peaks

  def time_of_extract(self, x_0, v_0):
    """Evaluates the time taken to extract a particle of mass M=1 and charge q=1 with
    initial velocity v_0 and position x_0 from the extraction plates."""
    self.Es = self.Vs / (2 * self.s)
    self.Ed = self.Vd / self.d

    coef = 1.02
    s0 = (self.s - x_0)
    a_s = self.Es
    if a_s > 1.e-25:
      Ts = coef / (a_s) * (-v_0 + np.sqrt(np.square(v_0) + 2 * a_s * s0))
    else:
      Ts = coef * s0 / v_0
    return Ts

  def time_of_flight(self, x_0, v_0):
    """Evaluates the time-of-flight of a particle of mass M=1 and charge q=1 with
    initial velocity v_0 and position x_0."""
    self.Es = self.Vs / (2 * self.s)
    self.Ed = self.Vd / self.d

    coef = 1.02
    s0 = (self.s - x_0)
    a_s = self.Es
    if a_s > 1.e-25:
      Ts = coef / (a_s) * (-v_0 + np.sqrt(np.square(v_0) + 2 * a_s * s0))
    else:
      Ts = coef * s0 / v_0

    v_s = v_0 + a_s * Ts
    a_d = self.Ed
    if a_d > 1.e-25:
      Td = coef / (a_d) * (-v_s + np.sqrt(np.square(v_s) + (2 * a_d * self.d)))
    else:
      Td = coef * self.d / v_s

    v_d = v_s + a_d * Td
    TD = coef * self.D / v_d
    return Ts + Td + TD

  def time_of_flight_bibliog(self, x_0, v_0):
    """Alternative evaluatation of time-of-flight of a particle of mass M=1
    and charge q=1 with initial velocity v_0 and position x_0 (from bibliography)."""
    q = 1.0                                 # WARNING: All have charge q=1.
    coef = 1.02  # *np.sqrt(2)

    U_0 = np.square(v_0) / 2.
    Us = U_0 + q * (self.s + x_0) * self.Es
    u = np.sqrt(Us + q * self.d * self.Ed)
    us = np.sqrt(Us)

    Ts = coef * (us - v_0) / (q * self.Es)
    Td = coef * (u - us) / (q * self.Ed)
    TD = coef * self.D / (2 * u)
    return Ts + Td + TD

  def data_to_array(self, unpack=False):
    """Conversion from signal to "numpy arrays", sorted by time

    Keyword Arguments:
    unpack -- (default False)
    """
    if self.times is not None:
      # especies ordenadas por masa (menores tiempos)
      ks = list(self.times.keys())
      ks.remove('time')
      ks.remove('signal')
      mint = np.argmax([self.times[k] for k in ks], axis=1)
      ks = [x for (y, x) in sorted(zip(mint, ks))]
      ks.insert(0, 'time')
      ks.append('signal')

      y = np.array([self.times[k] for k in ks])

      if unpack:
        return ks, y.T
      else:
        return ks, y
    return None, None

  def save_data(self, fname):
    """Save the signal to a data file"""
    if self.times is not None:
      ks, y = self.data_to_array()
      header = '  ' + '      '.join(ks)
      np.savetxt(fname, y.T, fmt='%.5e', header=header)

  def make_plot(self, especies=None, fname=None, **kvars):
    """Plot the signals in a "standard" form.

    Parameters
    ----------
    especies : dict
      masses to plot. If no present uses simulated fragments
    fname : string or None (default None)
      if not None -> Save the figure to "fname"
    kvars : Optional
      - negative = `True/False`             : if `True` -> Negative signal
      - graph_all=`True/False` (or show_all): if `True` -> Plot individual species
      - show_legend=`True/False`            : if `True` -> Show the legend

    """
    try:
      from matplotlib.pyplot import figure
    except BaseException:
      print("Plotting not available. Install matplotlib")
      return None

    if especies is None:
      especies = self.fragments
    negative = kvars.get('negative', False)
    graph_all = kvars.get('graph_all', kvars.get('show_all', False))
    if 'show_legend' in kvars:
      show_legend = kvars['show_legend']
    else:
      show_legend = not kvars.get('hide_legend', False)

    fig = figure(num='tof', figsize=(11, 7))
    fig.clf()
    ax = fig.add_subplot(111)
    if negative:
      signo = -1.
    else:
      signo = 1.

    Labels = []
    for k, p in sorted(list(especies.items()), key=lambda x: x[1]['M']):
      l = p.get('L', k)
      pico = self.times[k].argmax()

      Labels.append((l, pico))
      if graph_all:
        ax.plot(self.times['time'], signo * self.times[k], lw=3, label=l)

    y = signo * np.array(self.times['signal'])

    if show_legend and not graph_all:
      ax.plot(self.times['time'], y, '-ok', markersize=2)

      xlims = (max(self.times['time'].min(), 0),
               self.times['time'].max())
      Deltax = (xlims[1] - xlims[0]) / 20.
      ylims = ax.get_ylim()
      Deltay = (ylims[1] - ylims[0]) / 30.
      ax.set_ylim(ylims[0] - Deltay, ylims[1] + Deltay)
      ax.set_xlim(xlims[0] - Deltax, xlims[1] + Deltax)
      legbox = dict(facecolor='red', alpha=0.5, linewidth=2)
      dx = 1
      for l, p in Labels:
        xt = self.times['time'][p] + dx * Deltax
        yt = y[p] - Deltay
        ax.text(xt, yt, l,
                horizontalalignment='center', bbox=legbox, fontsize='large')
        dx = -dx
    else:
      ax.plot(self.times['time'], y, '-ok', markersize=2)

    ax.set_xlabel(r" $t(\mu s)$", fontsize="x-large")
    ax.set_ylabel(r"$\rho(t)$", fontsize='x-large')
    ncol = (len(list(especies.keys())) - 2) // 6 + 1
    if show_legend:
      ax.legend(loc='best', title='$T={0}^{{\circ}}K$'.format(
          self.Temperature), framealpha=0.5, ncol=ncol)
    legtitle = 'Parámetros del TOF\n'
    for k in self.tof_parameters:
      legtitle += r'${0}={1}$  '.format(k, self.__dict__[k])
    ax.set_title(legtitle)
    if fname is not None:
      fig.savefig(fname, bbox_inches='tight')
    return fig


class Peaks(dict):
  """Simple object describing spectra peaks for the TOF.

  It is essentially a dictionary, with a few added convenience methods

  Each peak is described by a dictionary whose key is the label of the fragment.

  The values are:

    - 'index': tuple
         has the form (im, i0, ip) with the indexes of the peak (i0) and the two positions where the width is obtained.
    - 'position': float
         is value of the coordinate where the peak is located, in microseconds.
    - 'height': float
         is the value of the peak.
    - 'width': float
         is the width of the peak in microseconds.
  """

  def __init__(self):
    self.headers = ['index', 'position', 'height', 'width']

  def tolist(self):
    """Returns a list with the data describing the peaks. The form is:
    ['Substance', 'index', 'position', 'height', 'width']
    """
    peaks = [[k] + [v[h] for h in self.headers] for k, v in self.items()]
    return peaks

  def __str__(self):
    """Nicer printing of peak information
    """
    L = self.tolist()
    headers = ['Substance'] + self.headers
    try:
      from tabulate import tabulate
      return "\n" + tabulate(L, headers=headers, tablefmt='simple')
    except BaseException:
      s = " ".join([h.center(15) for h in headers])
      s += "\n" + " ".join([15 * "-" for h in header]) + "\n"
      for r in L:
        s1 = [str(c).center(15) for c in r[:2]]
        s1 += ["{:11.6f}".format(c).center(15) for c in r[2:]]
        s += " ".join(s1) + "\n"
      return s



# -*- coding: utf-8 -*-
#
# This file is part of tof-simulator.
#
# tof-simulator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tof-simulator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tof-simulator.  If not, see <https://www.gnu.org/licenses/>.


import os
import re
import itertools
import collections as col
import numpy as np
from sys import stdout
from configparser import SafeConfigParser


try:
  datafile = os.path.join(os.path.dirname(os.path.realpath(__file__)), "nist_data.txt")
except BaseException:
  datafile = 'nist_data.txt'

if not os.path.exists(datafile):
  print("data file {} not found. Trying ...".format(datafile))
  datafile = os.path.join(os.getcwd(), "nist_data.txt")
if not os.path.exists(datafile):
  print("data file {} not found".format(datafile))

__NIST_ELEMENTS__ = {}


# def _is_scalar_or_string(val):
#   """Return whether the given object is a scalar or string like.
#   *copied from matplotlib.cbook*
#   """
#   return isinstance(val, str) or not np.iterable(val)


def _flatten(seq):
  """
  *Copied (and simplified for this use) from matplotlib.cbook*

  Return a generator of flattened nested containers.

  By: Composite of Holger Krekel and Luther Blissett
  From: https://code.activestate.com/recipes/121294/
  and Recipe 1.12 in cookbook
  """
  for item in seq:
    if (isinstance(item, str) or not np.iterable(item)) or item is None:
      yield item
    else:
      yield from _flatten(item)


def read_nist_data(threshold=0):
  """
  Read data from file *datafile* and fills the dictionary
  `__NIST_ELEMENTS__` with all the data

  Parameters
  ----------
  threshold: float
    Add all elements whose abundance is higher than threshold

  """
  global __NIST_ELEMENTS__
  __NIST_ELEMENTS__ = {}

  abbrev = ['Z', 'S', 'A', 'M', 'P', 'W']

  fi = open(datafile)
  end = False
  while not end:
    m = {}
    for f in abbrev:
      line = fi.readline()
      if line == '':
        end = True
        break
      v = line.split('=')[1].strip()
      if v == '': v = '0'
      m[f] = v

    if fi.readline().strip() != '': print('warning, se supone que estaba vacia')
    if end: break

    P = float(m['P'].split('(')[0])
    if P >= threshold:
      __NIST_ELEMENTS__[m['A'] + m['S']] = {
          'M': float(m['M'].split('(')[0]),
          'P': P,
          'S': m['S'],
          'A': int(m['A']),
          'Z': int(m['Z']),
          'L': make_label([(m['A'], m['S'], 0)], fmt='LaTeX'),
          'T': make_label([(m['A'], m['S'], 0)], fmt='LaTeXsimple')
      }
  fi.close()


def mass2conf(masses):
  """Format masses for a configuration file

  Parameters
  ----------
  masses: dict
    Information on substance(s)

  Returns
  -------
  str
    string with the format of a configuration file
  """
  s = ""
  for l, m in masses.items():
    s += "[{}]\n".format(l)
    for k, v in m.items():
      s += "{0}={1}\n".format(k, v)
    s += '\n'
  return s


def loadmass(fname):
  """ Load masses from configuration file

  Parameters
  ----------
  fname: str or Path()
    File to read masses from

  Returns
  -------
  dict
    a dictionary with the elements read
  """
  parser = SafeConfigParser(); parser.optionxform = str
  parser.read(fname)
  sections = parser.sections()
  m = {}
  for section_name in sections:
    m[section_name] = {}
    for name, value in parser.items(section_name):
      try:
        m[section_name][name] = float(value)
      except BaseException:
        m[section_name][name] = value
  return m


def make_label(ss, fmt='key'):
  """Construct the label for a given substance

  Parameters
  ----------
  ss: str
    Array of tuples [(A1,S1), (A2,S2), ...] for all atoms in the substance
  fmt: str
    Indicates the format of labels::

        'key'         -> "14N2"
        'Latex'       -> "$^{14}N_{2}$"
        'latexsimple' -> "$N_{2}$"
        'latexdoc'    -> r"\ce{^{14}N_{2}}"
        'mass'        -> "28"

  """
  ss.sort(key=lambda x: x[0])
  items = list(col.Counter(ss).items())
  items.sort(key=lambda x: ss.index(x[0]))
  if fmt.lower() == 'latexdoc':
    label = r'\ce{'
    for k, v in items:
      if v == 1:
        label += r' ^{{{0}}}{1}'.format(k[0], k[1])
      else:
        label += r' ^{{{0}}}{1}{2}'.format(k[0], k[1], v)

    if k[2] == 1: label += '^{+}'
    elif k[2] > 1: label += '^{{{0}+}}'.format(k[2])
    label += '}'

  elif fmt.lower() == 'latex':
    label = r'$'
    for k, v in items:
      if v == 1:
        label += r'^{{{0}}}\mathrm{{{1}}}'.format(k[0], k[1])
      else:
        label += r'^{{{0}}}\mathrm{{{1}}}_{{{2}}}'.format(k[0], k[1], v)
    if k[2] == 1: label += '^{+}'
    elif k[2] > 1: label += '^{{{0}+}}'.format(k[2])
    label += '$'

  elif fmt.lower() == 'latexsimple':
    label = r'$'
    for k, v in items:
      if v == 1:
        label += r'\mathrm{{{0}}}'.format(k[1], v)
      else:
        label += r'\mathrm{{{0}}}_{{{1}}}'.format(k[1], v)
    if k[2] == 1: label += '^{+}'
    elif k[2] > 1: label += '^{{{0}+}}'.format(k[2])
    label += '$'

  elif fmt.lower() == 'mass':
    label = '{}'.format(np.sum([int(x[0]) for x in ss]))

  else:
    ll = []
    for k, v in items:
      if v == 1:
        ll.append('{0}{1}'.format(k[0], k[1]))
      else:
        ll.append('{0}{1}{2}'.format(k[0], k[1], v))

    label = '-'.join(ll)
    if k[2] == 1:
      label += r'^{+}'
    else:
      label += r'^{{{0}+}}'.format(k[2])

  return label


def _splitelement(subst):
  """Analiza una substancia y retorna los componentes y su multiplicidad como una lista de pares

  Parameters
  ----------
  subst: str
    Substance to analyze

  For instance for CHCO2
  gives:  [('C',2), ('H',1),('O',2)]
  """

  comp = list([_f for _f in re.split(r"([A-Z][a-z]*\d*)", subst) if _f])
  componentes = col.OrderedDict()
  for k in comp:
    elemx = list([_f for _f in re.split("([0-9]+)", k) if _f])
    if len(elemx) == 1:
      elem, X = elemx[0], 1
    else:
      elem = elemx[0]; X = int(elemx[1])
    if elem in componentes: componentes[elem] += X
    else: componentes[elem] = X
  return list(componentes.items())


def _splitcharge(subst):
  """split element and charge
  """
  if '^' in subst:
    subst, q = subst.split('^')

    q = q.strip('{}')
    if len(q.strip('+')) < 1:
      carga = len(q)
    else:
      carga = int(q.strip('+'))
  else:
    carga = 1

  return subst.strip(), carga


def analyze_substance(subst, threshold=1.e-4, fragments=False, isotopes=True):
  """Devuelve todas las combinaciones con posibles isotopos (y su poblacion)
  de la sustancia con la formula dada.

  Parameters
  ----------
  subst: str
    substance to include (H2O, CO, SF6, N2, ...)
  threshold: float
    Minimum abundance that has to have an isotope to be included
  fragments: bool
    If True includes all fragments. For instance from 'CO' -> 'CO', 'C', 'O'
  isotopes: bool
    If True includes all isotope combinations


  Examples::

      analyze_substance('H2O').keys() =  ['1H1-2H1-16O1', '1H2-17O1', '1H2-16O1', '1H2-18O1']
      analyze_substance('CO').keys() =
             ['13C1-17O1', '13C1-18O1', '12C1-16O1', '12C1-18O1', '13C1-16O1', '12C1-17O1']
  """

  subst, carga = _splitcharge(subst)
  componentes = _splitelement(subst)

  elementos = {}
  multiplicidad = {}
  combinaciones = []
  for e, x in componentes:
    elem = _get_element(e)
    if not isotopes:                        # No analizamos composición isotópica
      mayor = max(elem, key=lambda x: elem[x]['P'])   # Isotopo de mayor población
      elem = {mayor: elem[mayor]}

    m = list(elem.keys())
    elementos.update(elem)

    aa = list(itertools.combinations_with_replacement(m, r=x))
    combinaciones.append(aa)
    for k in aa:
      multiplicidad[k] = len(set(itertools.permutations(k)))

  sustancia = {}
  for k in itertools.product(*combinaciones):
    ll = [(elementos[s]['A'], elementos[s]['S'], carga) for s in _flatten(k)]
    label = make_label(ll)
    P = np.product(np.array([elementos[s]['P'] for s in _flatten(k)]))
    P *= np.product(np.array([multiplicidad[k[j]] for j in range(len(k))]))

    sustancia[label] = {'S': subst,
                        'M': np.sum(np.array([elementos[s]['M'] for s in _flatten(k)])),
                        'L': make_label(ll, fmt='LaTeX'),
                        'T': make_label(ll, fmt='LaTeXSimple'),
                        'P': P,
                        'A': np.sum(np.array([int(elementos[s]['A']) for s in _flatten(k)])),
                        'Z': np.sum(np.array([int(elementos[s]['Z']) for s in _flatten(k)])),
                        'q': carga
                        }
  for k, v in list(sustancia.items())[:]:
    if v['P'] < threshold:
      sustancia.pop(k)
    else:
      v['P'] *= 100.
  return sustancia


def _get_element(*elems):
  """Devuelve uno o varios elementos (con todos sus isotopos)
  en la forma adecuada para el TOF
  Keyword Arguments:
  elems -- Elemento o elementos a devolver: por ejemplo H, Kr, N, Ne, ...
  """
  global __NIST_ELEMENTS__
  if __NIST_ELEMENTS__ == {}: read_nist_data()

  sp = {}
  for elem in elems:
    for k, v in list(__NIST_ELEMENTS__.items()):
      if v['S'] == elem:             # IMPORTANTE:
        sp[k] = v.copy()              # sin copy() hace una referencia a __NIST_ELEMENTS[k]
  return sp


def format_listmass(rows, end='\n', sep=' ', header='', footer=''):
  """Simple formatting of a list of substances. It is intended to use with the
  output of `get_lista()`

  Parameters
  ----------
  rows: list
    List where each row is a list with data on a single element
  end: str
    End of line string
  sep: str
    string used as separator within a line
  header: str
    Header to add before the data
  footer: str
    Footer to add after the data

  """

  s = header + '\n'

  rows.sort(key=lambda x: x[1])
  for r in rows:
    for c in r:
      s += "{}{}".format(c, sep)
    s = s[:-len(sep)] + end
  s += footer
  return s


class Sustancias(dict):
  """Object describing a list of substances.

  Examples:

      >>> elementos = ['N2', 'H2O', 'Ar^{3+}', 'Ar^{+}', 'Ar^2+', 'UF6']
      >>> m = Sustancias(elementos, threshold=1.e-3)
      >>> print(m)
      l                     M          P
      -------------  --------  ---------
      1H2-16O^{+}     18.0106  99.7341
      1H2-18O^{+}     20.0148   0.204953
      14N2^{+}        28.0061  99.2733
      14N-15N^{+}     29.0032   0.72535
      36Ar^{3+}       35.9675   0.3365
      36Ar^{+}        35.9675   0.3365
      36Ar^{2+}       35.9675   0.3365
      40Ar^{3+}       39.9624  99.6003
      40Ar^{+}        39.9624  99.6003
      40Ar^{2+}       39.9624  99.6003
      19F6-235U^{+}  349.034    0.7204
      19F6-238U^{+}  352.041   99.2742

  To get help use:

      >>> m.show_help()
      Create an object Sustancias as:
      #
        Sustancias(<elemento o sustancia>)
      #
      Fields in mass have the meaning:
      #
       Z:  str  Atomic Number
       S:  str  Atomic Symbol
       A:  str  Mass Number
       M:  str  Relative Atomic Mass
       P:  str  Isotopic Composition
       W:  str  Standard Atomic Weight
       q:  str  Charge

  """

  _columns = 'ALMPSZ'

  _fields = [('Z', 'Atomic Number'), ('S', 'Atomic Symbol'),
             ('A', 'Mass Number'),
             ('M', 'Relative Atomic Mass'),
             ('P', 'Isotopic Composition'),
             ('W', 'Standard Atomic Weight'),
             ('q', 'Charge')
             ]

  def __init__(self, elementos='', threshold=1.e-4, isotopes=True):
    """Initialize a `Sustancias` instance

    Parameters
    ----------
    self: type
      description
    elementos: str
      A string defining one or more substances
    threshold: float
      Threshold abundace for adding a given isotope
    isotopes: bool
      If `True` include all isotopes, otherwise only the most abundant.

    """
    self.thr = threshold
    self.iso = isotopes
    self.ListItems = []
    self._sortorder = 'M'
    self.add(elementos)

  def copy(self):
    """Copy the object and returns a new one."""
    p = Sustancias()
    p.thr = self.thr
    p.iso = self.iso
    p.ListItems = self.ListItems
    p._sortorder = self._sortorder
    p.update(self)
    return p

  def show_help(self):
    """Show help on use of Object Sustancias    """
    s = """Create an object Sustancias as:

  Sustancias(<elemento o sustancia>)


  Fields in mass have the meaning:

    """
    # s += (32) * '=' + '\n'
    for k, v in self._fields:
      s += '  {0}:  str  {1}\n'.format(k, v)
    print(s)

  def add(self, sustancias):
    """Returns a dictionary with all the information on the substances added.

    Parameters
    ----------
    sustancias: str
      Comma separated list of substances to be added, por instance:

          m = add('H20,O2,N2,UF5,UF6')

    """
    m = {}
    if isinstance(sustancias, str):
      elems = sustancias.split(',')
    else:
      elems = sustancias

    for e in elems:
      if e != '':
        m.update(analyze_substance(e.strip(), threshold=self.thr, isotopes=self.iso))
    self.update(m)
    self.ListItems.extend(list(m.keys()))
    self.sort()

  def remove(self, sustancias, criteria='fragment'):
    """Remove a fragment

    Parameters
    ----------
    sustancias: str or list of strings
      substance or fragment to remove
    criteria: str
      either 'fragment' (the default value) or 'substance':


    For the choices of 'criteria', we use:

    - 'fragment': e.g:  Ar^{2+}, SF6^++,  Removes exactly the fragment given
    - 'substance': e.g:  Ar, SF6, Xe, Remove all fragments with independently of the charge
    """

    if criteria == 'key':
      for_removal = sustancias
    else:
      if isinstance(sustancias, str):
        elems = sustancias.split(',')
      else:
        elems = sustancias
      for e in elems:
        for_removal = []

        if e.strip() != '':
          ee, q = _splitcharge(e)
          for k in self:          # Listamos los que vamos a remover
            if criteria == 'substance':
              remover = (ee == self[k]['S'])
            else:
              remover = (ee == self[k]['S'] and q == self[k]['q'])
            if remover:
              for_removal.append(k)

    for k in for_removal:  # Ahora lo removemos
      self.ListItems.remove(k)
      self.pop(k)

  def list(self, key=None):
    """Return a list of elements, optionally filtered with a criteria given by key

    Parameters
    ----------
    key: None or function
      - if key=None returns all elements
      - otherwise key has to be a function returning `True` or `False`.


    Examples:

         >>> list(key = lambda x:  x['P'] > 0.40)  # returns True if population greater than 40%
         >>> list(key = lambda x:  x['M'] > 50)    # returns True if Mass smaller than than 50 AMU
         >>> list(key = lambda x:  x['q'] > 2)     # returns True if charge greater than 2

    """
    if key is None: return self.ListItems
    return [k for k in self.ListItems if key(self[k])]
    # Lista = []
    # for k in self.ListItems:
    #   if key(self[k]): Lista.append(k)
    # return Lista

  def sort(self, order=None, reverse=False):
    """Sort the substances according to the specified order.
    The sort is stable.

    Parameters
    ----------
    self: type
      description
    order: str
      desired order. One of 'ALMPSZq'
    reverse: bool
      If `True` will be reversed

    """
    if order is None:
      order = self._sortorder
    else:
      self._sortorder = order
    if order not in self._columns + 'q':
      print('Sort order must be one of {}'.format(self._columns + 'q'))
      return

    lista = [(k, self[k]) for k in self.ListItems]  # Get in order
    lista.sort(key=lambda x: x[1][order], reverse=reverse)
    self.ListItems = [ll[0] for ll in lista]

  def saveconf(self, fname=None):
    """Save the substances to a file in a format that may be later read

    Parameters
    ----------
    self: type
      description
    fname: None or str or pathlib.Path()
      Name of the file. If `None` prints to screen

    """
    s = mass2conf(self)
    if fname is not None:
      with open(fname, 'w') as fo:
        fo.write(s)
    else:
      stdout.write(s)

  def loadconf(self, fname):
    """Load substances from file

    Parameters
    ----------
    fname: str or Path() object
      File with definition of substances to add


    File should have configuration format as::

        [34S1]
        S=S
        L=$^{34}\mathrm{S}$
        P=4.25
        M=33.9678669
        A=34
        Z=16

        [32S1]
        S=S
        L=$^{32}\mathrm{S}$
        P=94.99
        M=31.972071
        A=32
        Z=16

    """
    m = loadmass(fname)
    self.update(m)
    self.ListItems.extend(m.keys())
    self.sort()

  def get_lista(self, cols=None):
    """Return a list with the cols

    Parameters
    ----------
    cols: tuple/list or None
      A list of string selecting the fields to included, chosen from `"A,L,l,M,P,S,Z"`
    """

    lista = []
    if cols is None:
      columnas = self._columns
    else:
      columnas = cols

    for k in self.ListItems:
      a = []
      for x in columnas:
        if x == 'l':
          a.append(k)
        elif x in self[k]:
          a.append(self[k][x])
      lista.append(a)
    return lista

  def __str__(self):
    return self.to_table(cols=['l', 'M', 'P'], tablefmt='simple').replace('\n', '\n  ')

  def to_text(self):
    """Convenience function to format as a simple table
    """
    cols = ['S', 'l', 'M', 'P']
    headers = u'Substance Formula Mass Abundance'.split()
    return self.to_table(cols=cols, headers=headers, tablefmt='fancy_grid')

  def to_table(self, cols=None, headers=(), tablefmt='simple', floatfmt='g',
               numalign='decimal', stralign='left', missingval=''):
    """Format the list of sustances in Table format.

    This is a wrapper around the tabulate function
    from the tabulate package: https://pypi.python.org/pypi/tabulate.
    Available formats are:
    fancy_grid, grid, plain, psql, rst, simple, tsv
    latex, latex_raw, latex_booktabs, mediawiki, orgtbl, pipe,
    html

    Options `tablefmt`, `floatfmt`, `numalign`, `stralign`, `missingval` are passed to package `tabulate`, if available.

    Parameters
    ----------
    cols: list
      A list of strings with the characteristics to include. Must be in ALMPSZ
    headers: str
      String to include before data
    tablefmt: str
      Format of table
    floatfmt: str
      Format used to write numbers
    numalign: str
      Alignment for numbers
    stralign: str
      Alignment for strings
    missingval: str
      Value to write when values are missing

    """
    if headers == ():
      if cols is None:
        headers = self._columns
      else:
        headers = cols
    l = self.get_lista(cols)
    try:
      from tabulate import tabulate
    except BaseException:
      print('Output tables depend on package tabulate.')
      print("Install it from: https://pypi.python.org/pypi/tabulate")
      s = format_listmass(l, header=' '.join(headers))
      return s

    return '\n' + tabulate(l, headers, tablefmt, floatfmt, numalign, stralign, missingval)

  def to_latex(self, cols=None, headers=None, standalone=False):
    """Convert to LaTeX table using tabulate package

    Parameters
    ----------
    cols: list
      Substance characteristics to include.
    headers: str
      String to write before data.
    standalone: bool
      If `True` make a standalone document that may be compiled by LaTeX.

    """
    if standalone:
      if len(self) < 32: fontsize = r'\large'
      else: fontsize = ''
      head = r"""\documentclass[a4paper,12pt]{article}
\usepackage[version=4]{mhchem}
\usepackage[table]{xcolor}
\usepackage[top=1cm, bottom=1.5cm]{geometry}

\begin{document}
\setlength{\tabcolsep}{10pt}
\renewcommand{\arraystretch}{1.3}
\thispagestyle{empty}

\rowcolors{1}{black!20}{white}
""" + "{}\n".format(fontsize)
      foot = r"""
\end{document}
"""
    else:
      head = ""
      foot = ""

    formato = 'latex_raw'
    if cols is None: cols = ['L', 'M', 'P']
    if headers is None: headers = cols
    return head + self.to_table(cols=cols, headers=headers, tablefmt=formato) + foot

  def to_pdf(self, cols=None, headers=None, latexcommand='pdflatex', output='tmp.pdf'):
    """Convert to pdf format using latex.

    Export to a latex file and compiles it.

    Parameters
    ----------
    self: type
      description
    cols: str
      Substance characteristics to include
    headers: str
      Description to include before data
    latexcommand: str
      command to compile to pdf
    output: str
      Name of the output file

    """
    import subprocess as sub
    tt = os.path.splitext(output)[0] + '.tex'
    with open(tt, mode='w') as texfile:
      # with tempfile.NamedTemporaryFile(mode='w', suffix='.tex') as texfile:
      s = self.to_latex(cols=cols, headers=headers, standalone=True)

      texfile.write(s)
      texfile.close()
      cmd = latexcommand + " " + texfile.name
      ret = sub.call(cmd, shell=True, stdout=sub.PIPE, stderr=sub.PIPE)
      ret = sub.call(cmd, shell=True, stdout=sub.PIPE, stderr=sub.PIPE)

      try:                      # Clean-up intermediate files
        sub.call('latexmk -c ' + texfile.name, shell=True, stdout=sub.PIPE, stderr=sub.PIPE)
      finally:
        pass
    return ret



#
# This file is part of tof-simulator.
#
# tof-simulator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tof-simulator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tof-simulator.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["tof", "tof_gtk", "nist_elem"]

from .nist_elem import Sustancias, analyze_substance, mass2conf
from .tof import ToF



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# File: setup.py
#
# Copyright (C) 2020 Juan Fiol
# Written by Juan Fiol <juanfiol@gmail.com>
import os
from distutils.core import setup
from tofsim import version


def read(*rnames):
  return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


long_description = (
    "\n" + read('README.txt') + '\n' + version.COPYRIGHT
)


def setup_package():
  setup(name="tof-simulator",
        version=version.VERSION,
        description="Time-of-Flight Mass spectrometry simulator",
        long_description=long_description,
        url="https://github.com/fiolj/tof-simulator",
        keywords="simulation,mass-spectrometry,TOF",
        author="Juan Fiol",
        author_email="juanfiol@gmail.com",
        packages=['tofsim'],
        package_data={'tofsim': ['tof_gtk.xml', 'tof.conf', 'tof_teoria.pdf', 'nist_data.txt'],
                      },
        data_files=[('share/applications', ['tof_gtk.desktop']),
                    ],
        scripts=['scripts/tof_gtk'],
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ]
        )


if __name__ == '__main__':
  print(version.VERSION)
  setup_package()
